import html
import os
from os.path import realpath, join, dirname
import sys
import time
import warnings
import webbrowser

import jinja2

from ..frontend_semver import DECKGL_SEMVER


def in_jupyter():
    try:
        ip = get_ipython()  # noqa

        return ip.has_trait("kernel")
    except NameError:
        return False


def convert_js_bool(py_bool):
    if type(py_bool) != bool:
        return py_bool
    return "true" if py_bool else "false"


in_google_colab = "google.colab" in sys.modules


TEMPLATES_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "./templates/")
j2_loader = jinja2.FileSystemLoader(TEMPLATES_PATH)
j2_env = jinja2.Environment(loader=j2_loader, trim_blocks=True)
CDN_URL = "https://cdn.jsdelivr.net/npm/@deck.gl/jupyter-widget@{}/dist/index.js".format(DECKGL_SEMVER)


def cdn_picker(offline=False):
    # Support hot-reloading
    dev_port = os.getenv("PYDECK_DEV_PORT")
    if dev_port:
        print("pydeck running in development mode, expecting @deck.gl/jupyter-widget served at {}".format(dev_port))
        return (
            "<script type='text/javascript' src='http://localhost:{dev_port}/dist/index.js'></script>\n"
            "<script type='text/javascript' src='http://localhost:{dev_port}/dist/index.js.map'></script>\n"
        ).format(dev_port=dev_port)
    if offline:
        RELPATH_TO_BUNDLE = "../nbextension/static/index.js"
        with open(join(dirname(__file__), RELPATH_TO_BUNDLE), "r", encoding="utf-8") as file:
            js = file.read()
        return "<script type='text/javascript'>{}</script>".format(js)

    return "<script src='{}'></script>".format(CDN_URL)


def render_json_to_html(
    json_input,
    mapbox_key=None,
    google_maps_key=None,
    tooltip=True,
    css_background_color=None,
    custom_libraries=None,
    offline=False,
):
    js = j2_env.get_template("index.j2")
    css = j2_env.get_template("style.j2")
    css_text = css.render(css_background_color=css_background_color)
    html_str = js.render(
        mapbox_key=mapbox_key,
        google_maps_key=google_maps_key,
        json_input=json_input,
        deckgl_jupyter_widget_bundle=cdn_picker(offline=offline),
        tooltip=convert_js_bool(tooltip),
        css_text=css_text,
        custom_libraries=custom_libraries,
    )
    return html_str


def display_html(filename):
    """Converts HTML into a temporary file and opens it in the system browser."""
    url = "file://{}".format(filename)
    # Hack to prevent blank page
    time.sleep(0.5)
    webbrowser.open(url)


def iframe_with_srcdoc(html_str, width="100%", height=500):
    if isinstance(width, str):
        width = f'"{width}"'
    srcdoc = html.escape(html_str)

    iframe = f"""
        <iframe
            width={width}
            height={height}
            frameborder="0"
            srcdoc="{srcdoc}"
        ></iframe>
    """

    from IPython.display import HTML  # noqa

    with warnings.catch_warnings():
        msg = "Consider using IPython.display.IFrame instead"
        warnings.filterwarnings("ignore", message=msg)
        return HTML(iframe)


def render_for_colab(html_str, iframe_height):
    from IPython.display import HTML, Javascript  # noqa

    js_height_snippet = f"google.colab.output.setIframeHeight({iframe_height}, true, {{minHeight: {iframe_height}}})"
    display(Javascript(js_height_snippet))  # noqa
    display(HTML(html_str))  # noqa


def deck_to_html(
    deck_json,
    mapbox_key=None,
    google_maps_key=None,
    filename=None,
    open_browser=False,
    notebook_display=None,
    css_background_color=None,
    iframe_height=500,
    iframe_width="100%",
    tooltip=True,
    custom_libraries=None,
    as_string=False,
    offline=False,
):
    """Converts deck.gl format JSON to an HTML page"""
    html_str = render_json_to_html(
        deck_json,
        mapbox_key=mapbox_key,
        google_maps_key=google_maps_key,
        tooltip=tooltip,
        css_background_color=css_background_color,
        custom_libraries=custom_libraries,
        offline=offline,
    )

    if filename:
        with open(filename, "w+", encoding="utf-8") as f:
            f.write(html_str)

        if open_browser:
            display_html(realpath(f.name))

    if notebook_display is None:
        notebook_display = in_jupyter()

    if notebook_display and in_google_colab:
        render_for_colab(html_str, iframe_height)
        return html_str
    elif not filename and as_string:
        return html_str
    elif notebook_display:
        return iframe_with_srcdoc(html_str, iframe_width, iframe_height)



import base64
import os
import pathlib
import re


from pydeck.types import String
from pydeck.types.base import PydeckType

# See https://github.com/django/django/blob/stable/1.3.x/django/core/validators.py#L45
valid_url_regex = re.compile(
    r"^(?:http|ftp)s?://"
    r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"
    r"localhost|"
    r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
    r"(?::\d+)?"
    r"(?:/?|[/?]\S+)$",
    re.IGNORECASE,
)

valid_image_regex = re.compile(
    r".(gif|jpe?g|tiff?|png|webp|bmp)$",
    re.IGNORECASE,
)


def get_encoding(path: str) -> str:
    extension = pathlib.Path(path).suffix.replace(".", "")
    return f"data:image/{extension};base64,"


class Image(PydeckType):
    """Indicate an image for pydeck

    Parameters
    ----------

    path : str
        Path to image (either remote or local)
    """

    def __init__(self, path: str):
        if not self.validate(path):
            raise ValueError(f"{path} is not contain a valid image path")
        self.path = path
        self.is_local = not valid_url_regex.search(self.path)

    def __repr__(self):
        if self.is_local:
            with open(os.path.expanduser(self.path), "rb") as img_file:
                encoded_string = get_encoding(self.path) + base64.b64encode(img_file.read()).decode("utf-8")
                return repr(String(encoded_string, quote_type=""))
        else:
            return self.path

    def __eq__(self, other):
        return str(self) == str(other)

    @staticmethod
    def validate(path):
        # Necessary-but-not-sufficient checks for being a valid image for @deck.gl/json
        return any((valid_image_regex.search(path), valid_url_regex.search(path), path.startswith("data/image")))



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Uber Technologies, Inc.
# Distributed under the terms of the Modified BSD License.

from .bindings import map_styles, Deck, Layer, LightSettings, View, ViewState  # noqa

from .widget import DeckGLWidget  # noqa

from .nbextension import _jupyter_nbextension_paths  # noqa

from ._version import __version__  # noqa

from .settings import settings  # noqa



"""
Support serializing objects into JSON
"""
import json

from pydeck.types.base import PydeckType

# Attributes to ignore during JSON serialization
IGNORE_KEYS = [
    "mapbox_key",
    "google_maps_key",
    "deck_widget",
    "binary_data_sets",
    "_binary_data",
    "_kwargs",
]


def to_camel_case(snake_case):
    """Makes a snake case string into a camel case one

    Parameters
    -----------
    snake_case : str
        Snake-cased string (e.g., "snake_cased") to be converted to camel-case (e.g., "camelCase")

    Returns
    -------
    str
        Camel-cased (e.g., "camelCased") version of input string
    """
    output_str = ""
    should_upper_case = False
    for i, c in enumerate(snake_case):
        if c == "_" and i != 0:
            should_upper_case = True
            continue
        output_str = output_str + c.upper() if should_upper_case else output_str + c
        should_upper_case = False
    return output_str


def lower_first_letter(s):
    return s[:1].lower() + s[1:] if s else ""


def camel_and_lower(w):
    return lower_first_letter(to_camel_case(w))


def lower_camel_case_keys(attrs):
    """Makes all the keys in a dictionary camel-cased and lower-case

    Parameters
    ----------
    attrs : dict
        Dictionary for which all the keys should be converted to camel-case
    """
    for snake_key in list(attrs.keys()):
        if "_" not in snake_key:
            continue
        if snake_key == "_data":
            camel_key = "data"
        else:
            camel_key = camel_and_lower(snake_key)
        attrs[camel_key] = attrs.pop(snake_key)


def default_serialize(o, remap_function=lower_camel_case_keys):
    """Default method for rendering JSON from a dictionary"""
    if issubclass(type(o), PydeckType):
        return repr(o)
    attrs = vars(o)
    attrs = {k: v for k, v in attrs.items() if v is not None}
    for ignore_attr in IGNORE_KEYS:
        if attrs.get(ignore_attr):
            del attrs[ignore_attr]
    if remap_function:
        remap_function(attrs)
    return attrs


def serialize(serializable):
    """Takes a serializable object and JSONifies it"""
    return json.dumps(serializable, sort_keys=True, default=default_serialize)


class JSONMixin(object):
    def __repr__(self):
        """
        Override of string representation method to return a JSON-ified version of the
        Deck object.
        """
        return serialize(self)

    def to_json(self):
        """
        Return a JSON-ified version of the Deck object.
        """
        return serialize(self)



from .deck import Deck  # noqa
from .layer import Layer  # noqa
from .light_settings import LightSettings  # noqa
from .view import View  # noqa
from .view_state import ViewState  # noqa

from . import map_styles  # noqa
